#
# Copyright (c) 1998 Regents of the University of California.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. All advertising materials mentioning features or use of this software
#    must display the following acknowledgement:
#       This product includes software developed by the Computer Systems
#       Engineering Group at Lawrence Berkeley Laboratory.
# 4. Neither the name of the University nor of the Laboratory may be used
#    to endorse or promote products derived from this software without
#    specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# ------------
#
# Filename: tcl-import.tcl
#   -- Created on Tue May 19 1998
#   -- Author: Cynthia Romer <cromer@cs.berkeley.edu>
#

#
# Support for a simple, java-like import construct.
#
Class Import

#
# Wait to redefine the unknown proc until we are running in the interpreter,
# because the original unknown proc is not defined until after this tcl-object.tcl
# code is processed. (I believe the original unknown proc is actually defined when
# Tcl_AppInit calls Tcl_Init to set up the script library facility.)
#
proc redefine_unknown {} {
	#
	# If a proc/instproc is called on an unknown class, you'll wind up here.
	# If auto-importing is enabled, attempt to import the class.
	#
	rename unknown unknown.orig
	# Rather than redefining this procedure in tcl/library/init.tcl,
	# we can rename & augment it here for the mash interpreter.
	proc unknown { args } {
		# first try tcl's original unknown proc and return if successful
		if { ![catch "eval {unknown.orig} $args" m] } { 
			return
		}
		# otherwise, if autoimporting is enabled,
		# if able to import an item by this name, do so and return
		# btw, if the stuff in catch quotes causes "unknown" to get called, 
		# error = "too many nested calls to Tcl_EvalObj (infinite loop?)"
		if { ![catch "Import set autoimport_"] && [Import set autoimport_] } {
                        if { [catch "really_import [lindex $args 0]" import_error] } { 
				# if trying to import returns an error, puts error msgs for both
				error "$m \n$import_error"
			} 
		} else {
			# if not trying to import, puts original error msg
			error "$m" 
		}
	}

	# prevent this method from being called again
	proc redefine_unknown {} {}
}

#
# As new objects are needed, the unknown proc will catch them and
# import the files that define them and their methods.
# As an intended side-effect, explicit imports will be ignored
# until auto-import is disabled.
#
Import proc enable_autoimport {} {
	# XXX this should be done somewhere else
	import Class Object mashutils
	
	Import set autoimport_ 1
	redefine_unknown
	return
}


#
Import proc disable_autoimport {} {
	Import set autoimport_ 0
	return
}

#
# Auto-importing is disabled by default.
# (And because dynamic loading using unknown yet, due to the fact the unknown
#  isn't called when the -auperclass attribute is used in a Class defn.)
#
Import disable_autoimport 

#
# As long as the import procedure has not yet been invoked, the user is
# free to override mappings that may be read from importTables.
#
proc override_importTable_mapping { object file_list } {
	Import instvar instance_

	if ![info exists instance_] {
		set instance_ [new Import]
	}

	$instance_ override_importTable_mapping $object $file_list
}

#
# As long as the import procedure has not yet been invoked, the user is
# free to override mappings that may be read from importTables.
#
Import instproc override_importTable_mapping { object file_list } {
	$self instvar overrideTable_ table_made_

	if { [info exists table_made_] } {
		puts stderr "warning: ignoring \"override_importTable_mapping $object $file_list\" \n\
				It is illegal to modify the internal table_ after the first call to the import procedure."
		return
	}

	if { [info exists overrideTable_($object)] } {
		unset overrideTable_($object)
	}
	
	foreach file $file_list {
		set fname [condense_into_absolute_filename [ufile join [pwd] $file]]
		lappend overrideTable_($object) $fname
	}
}

#
# After the table_ has been generated, override existing mappings with
# any user-defined mappings.
#
Import instproc incorporate_table_overrides {} {
	$self instvar overrideTable_ table_

	foreach object [array names overrideTable_] {
		set table_($object) $overrideTable_($object)
	}
}

#
# When the Mash interpreter encounters "import <object(s)>", it will
# source the code for the supplied object(s).
#
# If import is explicitly called while we are configured to autoimport_,
# sorry, but we're gonna ignore that and wait for the class to be used before
# we import it.  On the other hand, if we are not configured to autoimport_,
# sure we'll import it for you right now.
#
proc import args {
	if { ![catch "Import set autoimport_"] && ![Import set autoimport_] } {
		if [catch "really_import $args" error_msg] {
			puts stderr "$error_msg"
		}
	}
}

#
# The first time this import procedure is called, an Import object is created
# and the importTables are read to create a mapping from objects to the
# files in which their class defn and method defns can be found.
#
proc really_import args {
	Import instvar instance_

	# Since importing objects can take awhile, especially if they are all imported
	# right at the very beginning, change the mouse-cursor to display a watch
	# so that, in MPlug, the user knows that the a mashlet is starting up.
	# If Yatin adds the ability to write to the status bar of your web browser,
	# here would be a great place to tell it to display "Importing..."
	# because seeing "Document: done" and nothing yet in your browser may lead
	# the user to assume the mashlet is not working.  Not that I ever thought that.  :-)
	global mash 
	if { $mash(environ) != "smash" } {
		set mouse_cursor [lindex [. config  -cursor] 4]
		. config  -cursor watch
	}
	
	if ![info exists instance_] {
		set instance_ [new Import]
	}

	set errormsg ""
	catch "eval $instance_ import $args" errormsg
	
	# return mouse-cursor to its normal state
	if { $mash(environ) != "smash" } {
		. config  -cursor $mouse_cursor
	}	

	if { [string length $errormsg] > 0 } {
		error $errormsg
	}
}

# Provide the value of an env variable and a list of the ':' separated items is returned.
#
# Since [split $env(TCLCL_IMPORT_DIRS) :]
# doesn't work very well if some of the ':' separated items are URLs (w/"http:"),
# you can use this proc to intelligently parse up an env variable
#
proc smart_parse_env_var { env_value } {
	set env $env_value
	while {[string length [set env [string trim $env ":"]]] != 0 } {
		if [regexp {^([^:]+)://([^:/]+)(:([0-9]+))?(/[^:]*)} $env url protocol server x port trailingpath] {
			lappend list $url
			regsub {([^:]+)://([^:/]+)(:([0-9]+))?(/[^:]*)} $env {} env
		} else {
			regexp {^[^:]+} $env dir
			lappend list $dir
			regsub {^[^:]+} $env {} env
		}
	}
	
	return $list
}

#
# Return a list of directories in which a readable importTable can be found.
# For every path in TCLCL_IMPORT_DIRS, importLocation will be read, if it exists, 
# and the absolute pathnames of the importTables that it points to will be appended to importTableList_ .
# For dirs in which a readable importLocation file is not found, the absolute pathname of the importTable in that dir, if one exists, will be appended instead.
# Directories may be named using complete pathnames or pathnames relative to CURRENTDIR.
#
Import instproc find_import_dirs {} {
	global env mash
	$self instvar tclcl_import_dirs_ import_table_list_

	# Generate a list of potential directories in which an importTable may be found
	foreach dir $tclcl_import_dirs_ {
		set importLocation [ufile join $dir importLocation]
		if [lindex [ufile readable $importLocation] 0] {
			set lines [read_file_into_array $importLocation]
			foreach line $lines {
				# append absolute filename on this line to the list
				lappend list [condense_into_absolute_filename [ufile join $dir $line]]
			}
		} else {
			lappend list $dir
		}
	}
	
	# prune the list down to the directories in which an importTable is actually readable
	set dirs ""
	set import_table_list_ ""
	if [info exists list] {
		foreach d $list {
			set import_table [ufile join $d importTable]
			if [lindex [ufile readable $import_table] 0] {
				lappend dirs $d
			}
			# record all of the importTables that we tried to read,
			# so can include in an error msg if no object mapping is found
			lappend import_table_list_ [condense_into_absolute_filename $import_table]
		}
	}

	if { [llength $dirs] > 0 } {
#		puts stderr "readable importTables found in: $dirs"
	} else {
#		puts stderr "no readable importTables found"
	}
	return $dirs
}

#
# Build an internal table of mappings from objects to the file(s) they are defined in.
#
Import instproc make_table {} {
	# retrieve a list of directories in which a readable importTable can be found
	set dirs [$self find_import_dirs]

	foreach d $dirs {
		$self read_dir $d
	}

	$self incorporate_table_overrides
}

#
# Returns a buffer read from the provided http <i>url</i>.
# If the url is unreadable, nothing is returned.
# To speed up performance, can send an <i>http_token</i> as the optional second arg if you know this url is readable.
#
proc read_url_into_buffer { url {http_token {}} } {
	if { $http_token == {} } {
		# rather than checking if { [ufile readable $url] != 1 }
		# just copying that code here to speed up performance	
		if [catch {set http_token [Http geturl $url]} m] {
			puts stderr "error executing \"::http::geturl $url\" = $m"
			return 
		} elseif {[lindex [set code [::http::code $http_token]] 1] != 200} {
			puts stderr "couldn't read file \"$url\": no such file or directory --> $code"
			return 
		}
	}
	
	set buffer [::http::data $http_token]
	return $buffer
}

#
# Reads a file (an http url is acceptable) and returns its contents as
# an array of lines.  The file should have already been verified to be
# readable.
#
proc read_file_into_array { file } {
	if [is_http_url $file] {
		set buffer [read_url_into_buffer $file]
		set lines [split $buffer "\n"]
	} else {
		set f [open $file "r"]
		while 1 {
			set line [gets $f]
			if [eof $f] {
				close $f
				break
			}
			lappend lines "$line"
		}
	}

	return $lines
}

#
# By reading the importTable in the provided directory <i>dir</i>,
# continue to define the elements of the table_ array.  Each element of
# this array is indexed using an object name and consists of a list of
# files (using absolute pathnames) in which the the object and its
# methods are defined.  Returns the table_ in list form.
#
Import instproc read_dir { dir } {
	$self instvar table_ indices_mapped_

	set importTableFile [condense_into_absolute_filename [ufile join $dir importTable]]

	# fetch the importTable and break it into a list of lines
	set lines [read_file_into_array $importTableFile]

	# for every line in the importTable, parse out the object and the filename,
	# adding the file to the appropriate list element of the table_ array
	# if it is not there already
	foreach line $lines {
		set index [lindex $line 0]
		#  use the absolute file name
		# (relative filenames are considered to be relative to the directory containing the importTable)
		set fname [condense_into_absolute_filename [ufile join $dir [lindex $line 1]]]

		# if a mapping for this object was already read from another importTable, ignore this one
		if [info exists indices_mapped_] {
			if [lsearch -exact $indices_mapped_ $index] {
				continue
			}
		}

		# if this object already has a mapping to this filename, skip it
		if {[info exists table_($index)]} {
			if {-1!=[lsearch -exact $table_($index) $fname]} {
				continue
			}
		}
		lappend table_($index) $fname
	}
	
	set indices_mapped_ [array names table_]

	return [array get table_]
}

#
# Map an OTcl class name <i>c</i> to a html-able name
#
proc class_to_file c {
	set f ""
	# map slashes to dashes
	set sep ""
	foreach s [split $c /] {
		set f "$f$sep$s"
		set sep -
	}
	return $f
}

#
# Return 1 if the <i>name</i> is a http URL and 0 otherwise.
# An http URL is of the form http://<server and possibly port num>/
# Note that the trailing slash is important.
#
proc is_http_url { name } {
	if [regexp {([^:]+)://([^:/]+)(:([0-9]+))?(/.*)} $name url protocol server x port trailingpath] {
		if { ![info exists protocol] } {
			return 0
		} else {
			return [regexp -nocase {http} $protocol]
		}
	} else {
		return 0
	}
}

#
# Add on a trailing / if not already there.
#
proc format_as_dir_string { dir_string } {
	return "[string trimright [ufile join $dir_string .] .]"
}

#
# the ufile proc is basically an enhanced file proc.
# The diff is that "ufile readable" & "ufile dirname" & "ufile join" procedures can handle URLs.
# The only difference from the file proc in behavior is that...
# for URLs, "file readable" returns a list:
#  if readable,   list = { 1 <http-token> }
#  if unreadable, list = { 0 <http-code or other error msg> }
#
# Note that "ufile join" continues to have the feature that an absolute pathname overrides
# any previous components.
#
proc ufile {option args} {
	if { $option == "readable" } {
		if { [llength $args] == 0 } {
			puts stderr "wrong # args: should be \"ufile readable name ?arg ...?\""
			return
		} else {
			set name [lindex $args 0]
			if [is_http_url $name] {
				set url $name
				if [catch {set token [Http geturl $url]} m] {
					return " 0 {error executing \"::http::geturl $url\" = $m} "
				} else {
					set code [::http::code $token]				
					if {[lindex $code 1] != 200} {
						return " 0 $code "
					} else {
						return " 1 $token "
					}
				}
			} else {
				eval {file $option} $args
			} 
		}
	} elseif { $option == "dirname" } {
		if { [llength $args] == 0 } {
			puts stderr "wrong # args: should be \"ufile dirname name ?arg ...?\""
			return
		} else {
			set name [lindex $args 0]
			if [is_http_url $name] {
				set url $name
				regexp {([^:]+)://([^:/]+)(:([0-9]+))?(/.*)} $name url protocol server x port trailingpath
				if {[string length $trailingpath] == 0} {
					set trailingpath /
				}
				set trailingpath [file dirname "$trailingpath"]
				return "$protocol://$server$x$trailingpath"
			} else {
				eval {file $option} $args
			} 			
		}
	} elseif { $option == "join" } {
		if { [llength $args] == 0 } {
			puts stderr "wrong # args: should be \"ufile join name ?arg ...?\""
			return
		} else {
			set base_url "[string trimright [lindex $args 0] /]/"
			set file_name [lindex $args 1]
			if [is_http_url $file_name] {
				return $file_name
			}
			if { [is_http_url $base_url] && [llength $args] == 2 } {
				# parse URL into components
				regexp {([^:]+)://([^:/]+)(:([0-9]+))?(/.*)} $base_url url protocol server x port trailingpath
				# get rid of initial ./ in file name
				regsub -all {^\./} $file_name {} file_name
				# change any /./ to / in file name
				regsub -all {/\./} $file_name {/} file_name
				# get rid of initial ../ in file name
				set counter 0
				while [regsub {^\.\./} $file_name {} file_name] {
					incr counter
				}
				# for each inital ../ removed, traverse up directory tree one level
				while { $counter > 0 } {
					set trailingpath [format_as_dir_string [ufile dirname $trailingpath]]
					incr counter -1
				}
				set trailingpath "[format_as_dir_string $trailingpath]$file_name"				
				return "$protocol://$server$x$trailingpath"
			} else {
				eval {file $option} $args
			} 			
		}
	} else {
		eval {file $option} $args
	}
}

#
# Eliminates the '..' and '.' in a pathname and returns the absolute path of an equivalent filename
# after following all symbolic links.
# This is useful if the filename was created using ufile join, which doesn't evaluate the '..' and '.'
# This can also be a useful pre-processing of a filename before comparing for equality.
# Note: this procedure is only valid for names of local files that are in executable directories.
# The filename is returned if the directory is invalid or not executable.
#
proc condense_into_absolute_filename { filename } { 
	set before_cd [pwd]
	# follow symlinks 
	while { ![catch "file readlink $filename"] } {
		set filename [file readlink $filename]
	}
	set dirname [ufile dirname $filename]
	set tailname [file tail $filename]
	set condensed_name $filename
	if { ![catch "cd $dirname"] } {
		set condensed_name [ufile join [pwd] $tailname]
	}
	cd $before_cd
	return $condensed_name
}

#
# Augment existing "source" procedure to get http files if <i>filename</i> starts with http.
#
rename source source.orig
proc source {fileName} {
	if [is_http_url $fileName] {
		set url $fileName
		set buffer [read_url_into_buffer $url]
		uplevel eval $buffer
	} else {
		uplevel source.orig $fileName
	}	

}

#
# Source the file by "uplevel \#0".
# If the file was previously sourced, it will not be sourced again. 
# Will fetch http files using url if <i>filename</i> starts with http.
# To speed up performance, can send an <i>http_token</i> as the optional second arg if you know this url is readable.
#
Import proc source { fileName {http_token {}} } {
	$self instvar loaded_ uniq_num_

	if { ![info exists uniq_num_] } {
		set uniq_num_ 0
	}

	# make sure the filename has been expanded to an absolute path
	# with sym links followed and extraneous '.'s and '..'s removed
	# so that we are most likely to recognize repeated attempts to source the same file
	set fileName [condense_into_absolute_filename $fileName]
	if [info exists loaded_($fileName)] {
#		puts stdout "$fileName previously sourced.  Not re-sourcing it."
		return 
	}
	set loaded_($fileName) 1

	# redirects explicit calls to the source proc within fileName
	# to this source proc in order to ensure that files
	# are not repeatedly sourced
	incr uniq_num_	
	uplevel \#0 "rename source source.$uniq_num_"
	uplevel \#0 {proc source {args} { Import source $args }}
	
	if [is_http_url $fileName] {
		set url $fileName
		set buffer [read_url_into_buffer $url $http_token]
		# catch errors so that errors in the file being sourced
		# do not cause this script to be exited prematurely
		# (btw, need {} around $buffer so variable substitutions occur in the global scope
		# and so blank lines at beginning of buffer don't cause errors
		if [catch "uplevel \#0 {$buffer}" errormsg] {
			puts "error in $fileName: $errormsg"
		}
	} else {
		# catch errors so that errors in the file being sourced
		# do not cause this script to be exited prematurely
		if [catch "uplevel \#0 source.orig $fileName" errormsg] {
			puts "error in $fileName: $errormsg"
		}
	}

	# disabling the redirection to this alternate source proc
	uplevel \#0 {rename source {}}
	uplevel \#0 "rename source.$uniq_num_ source"
	incr uniq_num_ -1
}

#
# We now require import mechanism across all mash scripts, so "Import
# enable proc" was removed.  Scripts that use "Import enable" won't flag
# an error because this just declares an object now.  But for this
# reason, best not to do anything in init.
#
Import instproc init {} {
}

#
# Read the environment variable, TCLCL_IMPORT_DIRS, and store the directories in a list instvar.
# Afterwards, makes a call to the instproc that generates the table.
#
Import instproc init_table {} {
	$self instvar tclcl_import_dirs_ table_made_
	global env

	# If TCLCL_IMPORT_DIRS is not set before first time import proc is called, it is set to '.'
	# Note that otherwise, '.' is not appended to TCLCL_IMPORT_DIRS.
	if { ![info exists env(TCLCL_IMPORT_DIRS)] } {
		set env(TCLCL_IMPORT_DIRS) .
	}

	foreach dir [smart_parse_env_var $env(TCLCL_IMPORT_DIRS)] {
		# If dir is relative, it is expanded to absolute.
		# Relative pathnames in TCLCL_IMPORT_DIRS will be considered relative to '.'
		# which is the directory the mash interpreter was launched from
		# (unless the cd proc has been called since then)
		# So if mashlets are bing run from a browser, '.' starts out as
		# the directory from which the browser was launched.
		lappend tclcl_import_dirs_ [condense_into_absolute_filename $dir]
	}

	# the first time import is called, build a table of mappings from objects to the file(s) they are defined in
	$self make_table
#	$self check_table
	set table_made_ 1
}

#
# Generates error messages after table_ has been created, notifying
# programmer if any files define multiple objects or if any objects are
# defined in multiple files, both of which are poor programming styles.
#
Import instproc check_table {} {
	$self instvar table_

	# check for multiple files per object
	foreach object [array names table_] {
		if { [llength $table_($object)] > 1} {
			puts stdout "Warning: $object is mapped to multiple files..."
			foreach file $table_($object) {
				puts stdout "     $file"
			}
			puts stdout "Good mash programming style is to import only one file per class."  		
			puts stdout "Redefining the class may make the alloc method unreachable and break superclass relationships."
		}
		foreach file $table_($object) {
			lappend inverse_table($file) $object
		}
	}

	# check for multiple objects per file
	foreach file [array names inverse_table] {
		if { [llength $inverse_table($file)] > 1} {
			puts stdout "Warning: multiple classes defined in $file..."
			foreach object $inverse_table($file) {
				puts stdout "     $object"
			}
			puts stdout "Good mash programming style is to define one class per file."
		}
	}
}

#
# Just verifies that the env(TCLCL_IMPORT_DIRS) hasn't been modified since last time import proc was called.
#
Import instproc consistency_check {} {
	global env
	$self instvar last_time_

	if { ![info exists last_time_] } {
		set last_time_ $env(TCLCL_IMPORT_DIRS)
	}
	
	if { $env(TCLCL_IMPORT_DIRS) != $last_time_ } {
		puts stderr "warning: ignoring modification to env(TCLCL_IMPORT_DIRS) \n\
				It is illegal to modify this after the first call to the import procedure."
	}
}

#
# Sources the file for this object.
# If file is unreadable, puts error msg.
# Returns 1 if file successfully sourced (now or previously)
# and 0 otherwise.
#
Import private source_file_4_object { file object } {	
	set file_readable_result [ufile readable $file]
	set file_readable [lindex $file_readable_result 0]                         
	if { $file_readable } {
		set http_token [lindex $file_readable_result 1] 
#		puts stdout "Importing $object from $file"
		Import source $file $http_token
		return 1
	} else {
		set errormsg [lindex $file_readable_result 1]    
		if { $errormsg != {} } {
#			puts stderr "*** error: can't get file $file for $object --> $errormsg"
		} else {
#			puts stderr "*** error: can't find file [condense_into_absolute_filename $file] for $object"
		}
		return 0 
	}
}
		
#
# For the objects provided as <i>args</i>, source the files in which their class & methods are defined.
# After attempting to import all supplied items, if any were unimportable,
# an error will be flagged with a detailed errormsg.
#
Import instproc import {args} {
	$self instvar table_ tclcl_import_dirs_ import_table_list_ table_made_

	if { ![info exists table_made_] } {
		$self init_table
	}

	$self consistency_check

	set errormsg ""
	foreach item $args {
		set file_list ""
		# generate a list of files to source in order to evaluate all the code for this object
		if [info exist table_($item)] {
			set file_list $table_($item)
			# although it's poor programming practice, an object can be defined in multiple files
			foreach file $file_list {
				$self source_file_4_object $file $item 
			}
		} else {
			# if the object is not in the table_, try searching for default name (<object>.mash)
			# in each of the env(TCLCL_IMPORT_DIRS)
			foreach dir $tclcl_import_dirs_ {
				lappend file_list [condense_into_absolute_filename [ufile join $dir [class_to_file $item].mash]]
			}
			set imported 0
			foreach file $file_list {
				# stop after find first readable default location
				if { [set imported [$self source_file_4_object $file $item]] } {
					break
				}
			}
			if { !$imported } {
				append errormsg "*** error: don't know how to import $item\n"
				append errormsg "    not mapped in: $import_table_list_\n"
				append errormsg "    and not found in default locations: $file_list\n"
			}
		}
	}
	if { [string length $errormsg] > 0 } {
		error $errormsg
	}
}
